@@ -1,3 +1,5 @@
+from urllib import urlencode
+from urlparse import urlparse, parse_qs
 import requests
 
 from .exceptions import FacepyError
@@ -19,62 +21,74 @@ class GraphAPI(object):
         self.session = requests.session()
         self.url = url.strip('/')
 
-    def get(self, path='', page=False, **options):
+    def get(self, path='', page=False, retry=3, **options):
         """
         Get an item from the Graph API.
 
         :param path: A string describing the path to the item.
         :param page: A boolean describing whether to return a generator that
                      iterates over each page of results.
+        :param retry: An integer describing how many times the request may be retried.
         :param options: Graph API parameters such as 'limit', 'offset' or 'since'.
 
         See `Facebook's Graph API documentation <http://developers.facebook.com/docs/reference/api/>`_
         for an exhaustive list of parameters.
         """
-
-        response = self._query('GET', path, options, page)
-
-        if isinstance(response, Exception):
-            raise response
+        response = self._query(
+            method = 'GET',
+            path = path,
+            data = options,
+            page = page,
+            retry = retry
+        )
 
         if response is False:
             raise self.FacebookError('Could not get "%s".' % path)
 
         return response
 
-    def post(self, path='', **data):
+    def post(self, path='', retry=0, **data):
         """
         Post an item to the Graph API.
 
         :param path: A string describing the path to the item.
-        :param options: Graph API parameters.
+        :param retry: An integer describing how many times the request may be retried.
+        :param data: Graph API parameters such as 'message' or 'source'.
 
         See `Facebook's Graph API documentation <http://developers.facebook.com/docs/reference/api/>`_
         for an exhaustive list of options.
         """
-
-        response = self._query('POST', path, data)
+        response = self._query(
+            method = 'POST',
+            path = path,
+            data = data,
+            retry = retry
+        )
 
         if response is False:
             raise self.FacebookError('Could not post to "%s"' % path)
 
         return response
 
-    def delete(self, path):
+    def delete(self, path, retry=3):
         """
         Delete an item in the Graph API.
 
         :param path: A string describing the path to the item.
+        :param retry: An integer describing how many times the request may be retried.
         """
-
-        response = self._query('DELETE', path)
+        response = self._query(
+            method = 'DELETE',
+            path = path,
+            retry = retry
+        )
 
         if response is False:
             raise self.FacebookError('Could not delete "%s"' % path)
 
         return response
 
-    def search(self, term, type, page=False, **options):
+    def search(self, term, type, page=False, retry=3, **options):
         """
         Search for an item in the Graph API.
 
@@ -82,6 +96,7 @@ class GraphAPI(object):
         :param type: A string describing the type of items to search for.
         :param page: A boolean describing whether to return a generator that
                      iterates over each page of results.
+        :param retry: An integer describing how many times the request may be retried.
         :param options: Graph API parameters, such as 'center' and 'distance'.
 
         Supported types are ``post``, ``user``, ``page``, ``event``, ``group``, ``place`` and ``checkin``.
@@ -89,7 +104,6 @@ class GraphAPI(object):
         See `Facebook's Graph API documentation <http://developers.facebook.com/docs/reference/api/>`_
         for an exhaustive list of options.
         """
-
         SUPPORTED_TYPES = ['post', 'user', 'page', 'event', 'group', 'place', 'checkin']
         if type not in SUPPORTED_TYPES:
             raise ValueError('Unsupported type "%s". Supported types are %s' % (type, ', '.join(SUPPORTED_TYPES)))
@@ -99,7 +113,7 @@ class GraphAPI(object):
             'type': type,
         }, **options)
 
-        response = self._query('GET', 'search', options, page)
+        response = self._query('GET', 'search', options, page, retry)
 
         return response
 
@@ -109,7 +123,7 @@ class GraphAPI(object):
 
         :param requests: A list of dictionaries with keys 'method', 'relative_url' and optionally 'body'.
 
-        Yields a list of responses and/or GraphAPI.FacebookError, GraphAPI.HTTPError instances.
+        Yields a list of responses and/or exceptions.
         """
         responses = self.post(
             batch = json.dumps(requests)
@@ -124,28 +138,29 @@ class GraphAPI(object):
                 yield None
                 continue
 
-            data = self._parse(response['body'])
+            try:
+                yield self._parse(response['body'])
+            except FacepyError as exception:
+                exception.request = request
+                yield exception
 
-            if isinstance(data, Exception):
-                data.request = request
-
-                yield data
-                continue
-
-            yield data
-
-    def fql(self, query):
+    def fql(self, query, retry=3):
         """
         Use FQL to powerfully extract data from Facebook.
 
         :param query: A FQL query or FQL multiquery ({'query_name': "query",...})
+        :param retry: An integer describing how many times the request may be retried.
 
         See `Facebook's FQL documentation <http://developers.facebook.com/docs/reference/fql/>`_
         for an exhaustive list of details.
         """
-        return self._query('GET', 'fql?%s' % urlencode({'q': query}))
+        return self._query(
+            method = 'GET',
+            path = 'fql?%s' % urlencode({'q': query}),
+            retry = retry
+        )
 
-    def _query(self, method, path, data=None, page=False):
+    def _query(self, method, path, data=None, page=False, retry=0):
         """
         Fetch an object from the Graph API and parse the output, returning a tuple where the first item
         is the object yielded by the Graph API and the second is the URL for the next page of results, or
@@ -159,28 +174,24 @@ class GraphAPI(object):
         data = data or {}
 
         def load(method, url, data):
-            if method in ['GET', 'DELETE']:
 
-                try:
+            try:
+                if method in ['GET', 'DELETE']:
                     response = self.session.request(method, url, params=data, allow_redirects=True)
-                except requests.RequestException as exception:
-                    raise self.HTTPError(exception.message)
 
+                if method in ['POST', 'PUT']:
+                    files = {}
 
-            if method in ['POST', 'PUT']:
-                files = {}
+                    for key in data:
+                        if hasattr(data[key], 'read'):
+                            files[key] = data[key]
 
-                for key in data:
-                    if hasattr(data[key], 'read'):
-                        files[key] = data[key]
+                    for key in files:
+                        data.pop(key)
 
-                for key in files:
-                    data.pop(key)
-
-                try:
                     response = self.session.request(method, url, data=data, files=files)
-                except requests.RequestException as exception:
-                    raise self.HTTPError(exception.message)
+            except requests.RequestException as exception:
+                raise self.HTTPError(exception.message)
 
             result = self._parse(response.content)
 
@@ -195,6 +206,13 @@ class GraphAPI(object):
             while url:
                 result, url = load(method, url, data)
 
+                # If number of results is smaller than the limit parameter passed, there's no need to query an extra page
+                qs = parse_qs(urlparse(url).query)
+                if 'limit' in qs:
+                    limit = int(qs['limit'][0])
+                    if len(result['data']) < limit:
+                        url = None
+
                 # Reset pagination parameters.
                 for key in ['offset', 'until', 'since']:
                     try:
@@ -206,8 +224,7 @@ class GraphAPI(object):
 
         # Convert option lists to comma-separated values.
         for key in data:
-            if (isinstance(data[key], list) or isinstance(data[key], tuple) or isinstance(data[key], set)) \
-                and all([isinstance(item, basestring) for item in data[key]]):
+            if isinstance(data[key], (list, set, tuple)) and all([isinstance(item, basestring) for item in data[key]]):
                 data[key] = ','.join(data[key])
 
         url = '%s/%s' % (self.url, path)
@@ -215,10 +232,16 @@ class GraphAPI(object):
         if self.oauth_token:
             data['access_token'] = self.oauth_token
 
-        if page:
-            return paginate(method, url, data)
-        else:
-            return load(method, url, data)[0]
+        try:
+            if page:
+                return paginate(method, url, data)
+            else:
+                return load(method, url, data)[0]
+        except FacepyError:
+            if retry:
+                return self._query(method, path, data, page, retry - 1)
+            else:
+                raise
 
     def _parse(self, data):
         """
@@ -226,7 +249,6 @@ class GraphAPI(object):
 
         :param data: A string describing the Graph API's response.
         """
-
         try:
             data = json.loads(data)
         except ValueError:
@@ -254,14 +276,14 @@ class GraphAPI(object):
                 else:
                     exception = self.FacebookError
 
-                return exception(
+                raise exception(
                     error.get('message'),
                     error.get('code', None)
                 )
 
             # Facebook occasionally reports errors in its legacy error format.
             if 'error_msg' in data:
-                return self.FacebookError(
+                raise self.FacebookError(
                     data.get('error_msg'),
                     data.get('error_code', None)
                 )
@@ -271,9 +293,8 @@ class GraphAPI(object):
     class FacebookError(FacepyError):
         """Exception for errors returned by the Graph API."""
 
-        def __init__(self, message, code):
-            super(GraphAPI.FacebookError, self).__init__(message)
-
+        def __init__(self, message=None, code=None):
+            self.message = message
             self.code = code
 
     class OAuthError(FacebookError):
